extern "C"
{
  #include <string.h>
  #include "wiring_private.h"
  #include "pins_arduino.h"
}
#include "FLISBus.h"

FLISBus::FLISBus(uint8_t pin)
{
  init(pin, 1);
}

FLISBus::FLISBus(uint8_t pin, uint8_t active)
{
  init(pin, active);
}

void FLISBus::send(uint8_t node, void *data)
{
  if(node == 0x00)
    return;
  uint8_t buffer[6], sum = 0;
  buffer[0] = node;
  memcpy(&buffer[1], data, 4);
  for(uint8_t i = 0; i < 5; i++)
    sum += buffer[i];
  buffer[5] = -sum;
  sendBytes(buffer, 6);
}

void FLISBus::send(uint8_t *address, void *data)
{
  uint8_t buffer[10], sum = 0;
  buffer[0] = 0x00;
  memcpy(&buffer[1], address, 4);
  memcpy(&buffer[5], data, 4);
  for(uint8_t i = 0; i < 9; i++)
    sum += buffer[i];
  buffer[9] = -sum;
  sendBytes(buffer, 10);
}

void FLISBus::sendNode(uint8_t domain, uint16_t subnet, uint8_t node, void *data)
{
  uint8_t address[4];
  address[0] = domain | 0x80;
  address[1] = (uint8_t)(subnet >> 8) | 0x80;
  address[2] = (uint8_t)(subnet >> 0);
  address[3] = node;
  send(address, data);
}

void FLISBus::sendGroup(uint8_t domain, uint8_t subnet, uint16_t group, void *data)
{
  uint8_t address[4];
  address[0] = domain | 0x80;
  address[1] = subnet & 0x7f;
  address[2] = (uint8_t)(group >> 8);
  address[3] = (uint8_t)(group >> 0);
  send(address, data);
}

void FLISBus::sendSerialNo(uint32_t serial, void *data)
{
  uint8_t address[4];
  address[0] = (uint8_t)(serial >> 24) & 0x7f;
  address[1] = (uint8_t)(serial >> 16);
  address[2] = (uint8_t)(serial >>  8);
  address[3] = (uint8_t)(serial >>  0);
  send(address, data);
}

void FLISBus::init(uint8_t pin, uint8_t active)
{
  _pin = pin;
  _active = active;
  digitalWrite(_pin, _active);
  pinMode(_pin, OUTPUT);
}

void FLISBus::sendByte(uint8_t data)
{
  digitalWrite(_pin, _active);
  uint8_t portmask = digitalPinToBitMask(_pin);
  volatile uint8_t *port = portOutputRegister(digitalPinToPort(_pin));
  for(uint8_t mask = 0x80; mask; mask >>= 1)
  {
    uint8_t oldSREG = SREG;
    cli();
    if(data & mask)
    {
      // 1
      *port ^= portmask;
      delayMicroseconds(10);
      *port ^= portmask;
    }
    else
    {
      // 0
      *port ^= portmask;
      delayMicroseconds(1);
      *port ^= portmask;
      delayMicroseconds(9);
    }
    SREG = oldSREG;
    delayMicroseconds(5);
  }
}

void FLISBus::sendBytes(uint8_t *data, uint8_t size)
{
  while(size--)
    sendByte(*data++);
  for(uint8_t i = 0; i < 10; i++)
    delayMicroseconds(10);
}

